#!/usr/bin/python

#
# For usage, see printUsage function or run with -h option.
#

import getopt
import asynchat
import asyncore
import math
import os
import re
import signal
import socket
import sys
import time

DEFAULT_PORT = 50431
PORT = DEFAULT_PORT
IS_DEBUG_OUTPUT_ENABLED = False
OPERATING_MODE_NORMAL = "normal"
OPERATING_MODE_FLUORESCENCE = "fluorescence"
OPERATING_MODE_TEST = "test"
OPERATING_MODES = [OPERATING_MODE_NORMAL, OPERATING_MODE_FLUORESCENCE,
  OPERATING_MODE_TEST]
DEFAULT_OPERATING_MODE = OPERATING_MODE_FLUORESCENCE
OPERATING_MODE = DEFAULT_OPERATING_MODE
OPERATING_MODE_FLUORESCENCE_RESET_TIMEOUT = 10.0
SE_K_EDGE_SCAN_ROI_START = 2326

def main(args):
  parseCommandLineOptions(args)
  ConnectionDispatcher(PORT)
  asyncore.loop()

def getProgramName(args=None):
  if args == None: args = sys.argv
  if len(args) == 0 or args[0] == "-c": return "PROGRAM_NAME"
  return os.path.basename(args[0])

def initializeSignalHandlers():
  signalToNameMap = {
    signal.SIGHUP:"SIGHUP",
    signal.SIGINT:"SIGINT",
    signal.SIGPIPE:"SIGPIPE",
    signal.SIGALRM:"SIGALRM",
    signal.SIGTERM:"SIGTERM",
    signal.SIGUSR1:"SIGUSR1",
    signal.SIGUSR2:"SIGUSR2"
  }

  def signalHandler(signum, frame):
    if signum in [signal.SIGINT, signal.SIGTERM]:
      print
      print "%s caught. Exiting." % signalToNameMap[signum]
      sys.exit(0)
    else:
      print >> sys.stderr
      print >> sys.stderr, "%s caught. Exiting." % signalToNameMap[signum]
      sys.exit(1)

  for each in signalToNameMap.keys():
    signal.signal(each, signalHandler)

def parseCommandLineOptions(args):
  (options, extra) = getopt.getopt(args[1:], "dp:m:h", ["debug", "port=",
    "mode=", "help"])

  if len(extra) > 0:
    print >> sys.stderr, "Error: unexpected command line option \"%s\"" % \
      extra[0]
    printUsage(sys.stderr)
    sys.exit(1)

  for eachOptName, eachOptValue in options:
    if eachOptName in ("-d", "--debug"):
      global IS_DEBUG_OUTPUT_ENABLED
      IS_DEBUG_OUTPUT_ENABLED = True
    elif eachOptName in ("-p", "--port"):
      global PORT
      PORT = int(eachOptValue)
    elif eachOptName in ("-m", "--mode"):
      if eachOptValue not in OPERATING_MODES:
        print >> sys.stderr, "Error: unknown mode \"%s\"" % eachOptValue
        printUsage(sys.stderr)
        sys.exit(1)
      global OPERATING_MODE
      OPERATING_MODE = eachOptValue
    elif eachOptName in ("-h", "--help"):
      printUsage(sys.stdout)
      sys.exit(0)

def printUsage(outStream):
  print >> outStream, """\
Usage: %s [OPTION]...
Options:
  -d,--debug        Print debug messages to stderr
  -p,--port=NUMBER  Listen on the specified port NUMBER for incoming
                    connections (default: %d)
  -m,--mode=STRING  Set the operating mode to STRING; STRING may be one of the
                    following:
                      normal
                        The spectrum will be based on a real acquisition of
                        100 seconds with a 1 kHz zero peak generator
                        frequency; the returned spectrum will be scaled
                        according to the acquisition real time
                      fluorescence
                        Simulate data for an Se K fluorescence scan by setting
                        all channel counts to 0 except channel %d (first
                        channel is 1) which is set to an Se K ROI sum and
                        channels 1-155 which contain a simulated zero peak;
                        with each successive acquisition, the Se K ROI sum is
                        adjusted to correspond to the next step in the scan;
                        the internal step position in the scan data will
                        automatically reset to the beginning if %.1f seconds
                        have passed since the last acquisition request
                      test
                        The spectrum will always consist of all sevens except
                        the first channel which will have the value 6 and the
                        last channel which will have the value 8
                    (default: %s)
  -h,--help         Print usage message and exit
""" % (getProgramName(), DEFAULT_PORT, SE_K_EDGE_SCAN_ROI_START + 1,
  OPERATING_MODE_FLUORESCENCE_RESET_TIMEOUT, DEFAULT_OPERATING_MODE)

class ConnectionDispatcher(asyncore.dispatcher):
  def __init__(self, port):
    asyncore.dispatcher.__init__ (self)
    self.port = port
    self.create_socket(socket.AF_INET, socket.SOCK_STREAM)
    self.set_reuse_addr()
    self.bind(("", port))
    self.listen(5)

  def handle_accept(self):
    ConnectionSession(self.accept())

class ConnectionSession(asynchat.async_chat):
  CommaOrSpaceDelimRe = re.compile(r" *, *| +(?!,)")

  THERMOSTAT_STATE_COOL_MAX = "-"
  THERMOSTAT_STATE_STABLE = "+"

  COOLING_MODE_PELTIER_OFF = 0
  COOLING_MODE_THERMOSTAT_OP = 1
  COOLING_MODE_MAX_CURRENT = 2

  # indexed by cooling mode, target temps in 1/10 deg C
  COOLING_MODE_TEMPS = [280, -210, -455]

  # in real life, it warms up more slowly than it cools down
  TEMP_CHANGE_RATE = 19.6 # 1/10 deg C per sec

  # indexed by cooling mode, target currents in mA
  COOLING_MODE_CURRENTS = [55, 765, 1120]

  def __init__(self, (conn, addr)):
    asynchat.async_chat.__init__(self, conn)
    self.set_terminator("\r")
    self.outputTerminator = "\r"
    self.buffer = ""
    if OPERATING_MODE == OPERATING_MODE_NORMAL:
      self.spectrum = self.createNormalSpectrum()
      self.spectrumAcquisitionTime = 100.0
    elif OPERATING_MODE == OPERATING_MODE_FLUORESCENCE:
      self.spectrum = self.createFluorescenceSpectrum()
      self.spectrumAcquisitionTime = 1.0
    elif OPERATING_MODE == OPERATING_MODE_TEST:
      self.spectrum = self.createTestSpectrum()
      self.spectrumAcquisitionTime = 1.0
    else:
      raise Exception("Programmer error: unknown OPERATING_MODE \"%s\"" % \
        OPERATING_MODE)
    self.currentAcquisitionRealTime = 0.0
    self.realTime = 0.0
    self.acquisitionTime = 0.0
    self.acquisitionStartTime = 0.0
    # must be set to the period of the self.spectrum data
    self.zeroPeakGeneratorPeriod = 100
    self.isAcquisitionPaused = False
    self.seKEdgeScanStepCount = 0
    self.thermostatState = self.THERMOSTAT_STATE_COOL_MAX
    self.coolingMode = self.COOLING_MODE_PELTIER_OFF
    self.coolingModeChangeTime = 0.0
    self.temperature = self.COOLING_MODE_TEMPS[self.coolingMode] # 1/10 deg C

  def collect_incoming_data(self, data):
    self.buffer = self.buffer + data

  def found_terminator(self):
    data = self.buffer
    self.buffer = ""
    self.handleClientRequest(data)

  def handleClientRequest(self, request):
    request = request.strip()
    if IS_DEBUG_OUTPUT_ENABLED: sys.stderr.write("< \"%s\"\n" % request)
    if request == "$CC":
      self.sendClientResponse("!CC")
    elif request == "$DF":
      self.sendClientResponse("!DF %d" % self.coolingMode)
    elif request.startswith("$DK "):
      try:
        value = int(request[4:])
        if value not in (0, 1, 2): raise ValueError("Invalid argument")
        self.coolingMode = value
        self.temperatureLastChangedTime = time.time()
        self.sendClientResponse("!DK")
      except ValueError, e:
        self.sendClientResponse("!ERROR: 07")
    elif request == "$DR":
      self.updateThermostatStateAndTemperature()
      self.sendClientResponse("!DR %s" % self.thermostatState)
    elif request == "$DT":
      self.updateThermostatStateAndTemperature()
      self.sendClientResponse("!DT %d" % self.temperature)
    elif request == "$FP":
      responseStatus = "+"
      if self.isAcquiring(): responseStatus = "-"
      self.sendClientResponse("!FP %s" % responseStatus)
    elif request.startswith("$MP "):
      if request[4:] in ("ON", "+"):
        if self.isAcquiring():
          self.currentAcquisitionRealTime += max(0.0, time.time() - \
              self.acquisitionStartTime)
          self.isAcquisitionPaused = True
        self.sendClientResponse("!MP")
      elif request[4:] in ("NO", "-"):
        if self.isAcquisitionPaused:
          self.acquisitionStartTime = time.time()
          self.isAcquisitionPaused = False
        self.sendClientResponse("!MP")
      else:
        self.sendClientResponse("!ERROR: 06")
    elif request == "$MS":
      self.sendClientResponse("!MS %d" % (int(self.realTime * 1000.0)))
    elif request.startswith("$MT "):
      if OPERATING_MODE == OPERATING_MODE_FLUORESCENCE:
        if time.time() - self.acquisitionStartTime > \
            OPERATING_MODE_FLUORESCENCE_RESET_TIMEOUT:
          self.resetSeKEdgeScan()
        else:
          self.stepSeKEdgeScan()
      self.realTime = 0.0
      self.currentAcquisitionRealTime = 0.0
      self.acquisitionTime = float(request[4:]) / 1000.0
      self.acquisitionStartTime = time.time()
      self.isAcquisitionPaused = False
      self.sendClientResponse("!MT")
    elif request == "$PI":
      self.sendClientResponse("!PI %d" % \
        self.COOLING_MODE_CURRENTS[self.coolingMode])
    elif request.startswith("$SM "):
      (format,mode) = self.CommaOrSpaceDelimRe.split(request[4:].strip())
      (format,mode) = (int(format),int(mode))
      if format == 4 and mode == 0:
        self.sendClientResponse("!SM")
      else:
        self.sendClientResponse("!ERROR: 06")
    elif request.startswith("$SS "):
      self.realTime = self.computeCurrentRealTime()
      self.sendClientResponse("!SS")
      parameters = self.CommaOrSpaceDelimRe.split(request[4:].strip())
      if len(parameters) == 3: parameters.append("4096")
      (start,stepWidth,sumWidth,maxdata) = parameters
      (start,stepWidth,sumWidth,maxdata) = \
        (int(start),int(stepWidth),int(sumWidth),int(maxdata))
      dataCount = 0
      for i in xrange(start, 4096, stepWidth):
        if dataCount >= maxdata: break
        value = 0
        for j in xrange(i, i + sumWidth): value += self.spectrum[j]
        value = int(float(value) * (self.realTime / \
          self.spectrumAcquisitionTime))
        self.send4ByteClientResponse(value)
        dataCount += 1
    elif request.startswith("$SR "):
      self.sendClientResponse("!SR")
      parameters = self.CommaOrSpaceDelimRe.split(request[4:].strip())
      if len(parameters) == 3: parameters.append("4096")
      (start,stepWidth,sumWidth,maxdata) = parameters
      (start,stepWidth,sumWidth,maxdata) = \
        (int(start),int(stepWidth),int(sumWidth),int(maxdata))
      dataCount = 0
      for i in xrange(start, 4096, stepWidth):
        if dataCount >= maxdata: break
        value = 0
        for j in xrange(i, i + sumWidth): value += self.spectrum[j]
        value = int(float(value) * (self.realTime / \
          self.spectrumAcquisitionTime))
        self.send4ByteClientResponse(value)
        dataCount += 1
    elif request == "$FN":
      self.sendClientResponse("!FN %d" % self.zeroPeakGeneratorPeriod)
    else:
      self.sendClientResponse("!ERROR: 01")

  def sendClientResponse(self, response):
    self.push(response + self.outputTerminator)
    if IS_DEBUG_OUTPUT_ENABLED: sys.stderr.write("> \"%s\"\n" % response)

  def send4ByteClientResponse(self, value):
    fourBytes = ""
    for offset in [24, 16, 8, 0]:
      fourBytes = fourBytes + chr((value >> offset) & 0xff)
    self.push(fourBytes)
    if IS_DEBUG_OUTPUT_ENABLED:
      sys.stderr.write("%02x %02x %02x %02x (%d)\n" % (ord(fourBytes[0]),
        ord(fourBytes[1]), ord(fourBytes[2]), ord(fourBytes[3]), value))

  def updateThermostatStateAndTemperature(self):
    if self.temperature == self.COOLING_MODE_TEMPS[self.coolingMode]: return
    if self.thermostatState != self.THERMOSTAT_STATE_COOL_MAX:
      self.thermostatState = self.THERMOSTAT_STATE_COOL_MAX
    timeNow = time.time()
    changeInTemperature = int((timeNow - self.temperatureLastChangedTime) * \
      self.TEMP_CHANGE_RATE)
    if self.temperature < self.COOLING_MODE_TEMPS[self.coolingMode]:
      self.temperature = min(
        self.COOLING_MODE_TEMPS[self.coolingMode],
        self.temperature + changeInTemperature)
    else:
      self.temperature = max(
        self.COOLING_MODE_TEMPS[self.coolingMode],
        self.temperature - changeInTemperature)
    if self.temperature == self.COOLING_MODE_TEMPS[self.coolingMode]:
      if self.coolingMode != self.COOLING_MODE_PELTIER_OFF:
        self.thermostatState = self.THERMOSTAT_STATE_STABLE
    self.temperatureLastChangedTime = timeNow

  def computeCurrentRealTime(self):
    if self.isAcquisitionPaused:
      return self.currentAcquisitionRealTime
    else:
      return min(self.acquisitionTime, time.time() - \
        self.acquisitionStartTime + self.currentAcquisitionRealTime)

  def isAcquiring(self):
    if self.isAcquisitionPaused: return False
    else: return self.computeCurrentRealTime() < self.acquisitionTime

  def createNormalSpectrum(self):
    return NORMAL_SPECTRUM[:]

  def createFluorescenceSpectrum(self):
    # a Gaussian function <http://en.wikipedia.org/wiki/Gaussian_function> to
    # generate a low dead time zero peak for a 1.0 second acquisition with a 1
    # kHz zero peak generator frequency
    def gaussianZeroPeak(x,a=150.0,b=75.5,c=2.5):
      return int(a * math.exp(-((x-b)**2) / (2 * (c**2))))
    spectrum = []
    for i in xrange(156): spectrum.append(gaussianZeroPeak(i))
    for i in xrange(156, 4096): spectrum.append(0)
    spectrum[SE_K_EDGE_SCAN_ROI_START] = SE_K_EDGE_SCAN_ROI_SUMS[0]
    return spectrum

  def createTestSpectrum(self):
    spectrum = []
    spectrum.append(6)
    for i in xrange(1, 4095): spectrum.append(7)
    spectrum.append(8)
    return spectrum

  def stepSeKEdgeScan(self):
    if IS_DEBUG_OUTPUT_ENABLED:
      sys.stderr.write("Stepping Se K edge scan; setting step count to %d\n" \
        % (self.seKEdgeScanStepCount + 1))
    self.seKEdgeScanStepCount += 1
    self.spectrum[SE_K_EDGE_SCAN_ROI_START] = SE_K_EDGE_SCAN_ROI_SUMS[
      min(self.seKEdgeScanStepCount, len(SE_K_EDGE_SCAN_ROI_SUMS) - 1)]

  def resetSeKEdgeScan(self):
    if IS_DEBUG_OUTPUT_ENABLED:
      sys.stderr.write("Resetting Se K edge scan; setting step count to 0\n")
    self.seKEdgeScanStepCount = 0
    self.spectrum[SE_K_EDGE_SCAN_ROI_START] = SE_K_EDGE_SCAN_ROI_SUMS[0]

# Data is the sum of channels [2327-2347] (the first channel is 1) for each
# Energy position in an Se K edge fluorescence scan with a 1.0 second
# acquisition time and a 1 kHz zero peak generator frequency.
SE_K_EDGE_SCAN_ROI_SUMS = [
  52, 59, 59, 66, 109, 149, 148, 202, 224, 222, 293, 374, 439, 596, 721, 728,
  788, 745, 652, 702, 684, 698, 698, 662, 603, 701, 658, 679, 657, 614, 631,
  626, 596, 613, 613, 581, 565, 592, 543, 545, 542, 527, 464, 527, 485, 484,
  489, 508, 431, 444, 485, 464, 470, 507, 460, 421, 381, 378, 336, 336, 320,
  336, 298, 317, 294, 280, 280, 275, 269, 257, 221
]

# Data is the channel counts for a 100.0 second acquisition with a 1 kHz zero
# peak generator frequency.
NORMAL_SPECTRUM = [
  0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0,
  0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0,
  0, 0, 0, 0, 0, 0, 1, 2, 2, 2, 7, 7, 8, 8, 36, 36, 36, 36, 114, 115, 115,
  115, 311, 311, 311, 311, 698, 698, 699, 699, 1340, 1340, 1340, 1340, 2172,
  2173, 2173, 2173, 3102, 3102, 3102, 3102, 3642, 3642, 3642, 3642, 3665,
  3665, 3665, 3665, 3149, 3149, 3150, 3150, 2309, 2310, 2310, 2310, 1426,
  1427, 1427, 1427, 755, 755, 755, 755, 313, 313, 313, 313, 113, 113, 113,
  113, 35, 36, 36, 36, 14, 15, 15, 15, 2, 3, 3, 3, 1, 1, 1, 1, 1, 1, 1, 1, 0,
  0, 0, 0, 0, 1, 1, 1, 0, 0, 1, 1, 0, 0, 0, 0, 0, 0, 0, 0, 28, 28, 28, 28,
  279, 279, 280, 280, 378, 378, 379, 379, 393, 393, 394, 394, 369, 370, 370,
  370, 337, 337, 337, 337, 343, 344, 344, 344, 343, 343, 343, 343, 331, 331,
  331, 331, 315, 315, 315, 315, 285, 285, 286, 286, 287, 287, 287, 287, 292,
  293, 293, 293, 291, 291, 291, 291, 280, 280, 280, 280, 275, 276, 276, 276,
  262, 262, 262, 262, 271, 271, 271, 271, 261, 261, 262, 262, 267, 268, 268,
  268, 267, 268, 268, 268, 254, 254, 254, 254, 252, 252, 253, 253, 236, 236,
  236, 236, 245, 246, 246, 246, 230, 230, 231, 231, 235, 236, 236, 236, 227,
  228, 228, 228, 222, 223, 223, 223, 242, 243, 243, 243, 220, 220, 221, 221,
  207, 207, 207, 207, 223, 223, 223, 223, 217, 217, 217, 217, 221, 221, 221,
  221, 200, 200, 200, 200, 205, 205, 205, 205, 194, 194, 194, 194, 213, 213,
  213, 213, 209, 209, 209, 209, 206, 206, 206, 206, 204, 205, 205, 205, 201,
  202, 202, 202, 185, 185, 185, 185, 206, 206, 206, 206, 199, 199, 199, 199,
  187, 187, 188, 188, 182, 182, 183, 183, 191, 191, 192, 192, 196, 197, 197,
  197, 188, 188, 188, 188, 184, 184, 184, 184, 199, 199, 200, 200, 198, 198,
  199, 199, 185, 185, 186, 186, 191, 191, 191, 191, 195, 195, 195, 195, 191,
  191, 191, 191, 196, 196, 197, 197, 191, 191, 192, 192, 197, 198, 198, 198,
  192, 192, 193, 193, 180, 181, 181, 181, 181, 182, 182, 182, 186, 186, 186,
  186, 194, 194, 194, 194, 191, 191, 191, 191, 205, 205, 206, 206, 199, 200,
  200, 200, 192, 192, 192, 192, 202, 202, 203, 203, 201, 201, 201, 201, 206,
  206, 206, 206, 213, 213, 213, 213, 226, 226, 226, 226, 250, 250, 251, 251,
  280, 281, 281, 281, 305, 305, 306, 306, 339, 339, 340, 340, 360, 360, 361,
  361, 363, 363, 364, 364, 351, 351, 352, 352, 296, 296, 296, 296, 265, 265,
  265, 265, 264, 264, 265, 265, 229, 229, 229, 229, 212, 212, 212, 212, 213,
  213, 213, 213, 216, 216, 216, 216, 203, 204, 204, 204, 211, 212, 212, 212,
  232, 233, 233, 233, 260, 260, 260, 260, 277, 277, 278, 278, 296, 296, 297,
  297, 295, 295, 295, 295, 290, 290, 290, 290, 286, 286, 287, 287, 256, 256,
  256, 256, 244, 244, 244, 244, 215, 216, 216, 216, 223, 223, 223, 223, 196,
  196, 197, 197, 214, 215, 215, 215, 185, 186, 186, 186, 197, 197, 198, 198,
  198, 198, 198, 198, 205, 205, 205, 205, 205, 206, 206, 206, 192, 192, 192,
  192, 181, 182, 182, 182, 205, 205, 206, 206, 218, 219, 219, 219, 206, 207,
  207, 207, 202, 202, 202, 202, 202, 202, 202, 202, 207, 208, 208, 208, 195,
  195, 195, 195, 208, 208, 209, 209, 217, 217, 217, 217, 202, 202, 203, 203,
  211, 211, 211, 211, 208, 208, 208, 208, 220, 220, 220, 220, 206, 206, 206,
  206, 208, 209, 209, 209, 223, 223, 224, 224, 227, 227, 227, 227, 234, 235,
  235, 235, 226, 226, 226, 226, 243, 243, 243, 243, 230, 231, 231, 231, 231,
  231, 231, 231, 245, 245, 246, 246, 248, 248, 248, 248, 267, 267, 268, 268,
  255, 255, 256, 256, 245, 246, 246, 246, 248, 248, 248, 248, 255, 255, 255,
  255, 257, 257, 258, 258, 266, 267, 267, 267, 279, 280, 280, 280, 291, 292,
  292, 292, 274, 274, 275, 275, 288, 289, 289, 289, 305, 305, 305, 305, 321,
  321, 321, 321, 332, 332, 333, 333, 328, 329, 329, 329, 350, 351, 351, 351,
  371, 371, 372, 372, 385, 386, 386, 386, 416, 416, 417, 417, 462, 463, 463,
  463, 514, 515, 515, 515, 604, 605, 605, 605, 796, 796, 796, 796, 1177, 1177,
  1177, 1177, 1823, 1824, 1824, 1824, 2994, 2995, 2995, 2995, 4918, 4918,
  4918, 4918, 7447, 7448, 7448, 7448, 10476, 10476, 10477, 10477, 13772,
  13772, 13773, 13773, 16179, 16180, 16180, 16180, 17701, 17701, 17701, 17701,
  17333, 17333, 17333, 17333, 15390, 15391, 15391, 15391, 12438, 12439, 12439,
  12439, 9227, 9227, 9227, 9227, 6272, 6273, 6273, 6273, 3936, 3936, 3937,
  3937, 2579, 2579, 2580, 2580, 1956, 1956, 1956, 1956, 1929, 1929, 1929,
  1929, 2574, 2574, 2575, 2575, 3793, 3793, 3793, 3793, 5517, 5518, 5518,
  5518, 7602, 7603, 7603, 7603, 9653, 9654, 9654, 9654, 11476, 11476, 11476,
  11476, 12262, 12263, 12263, 12263, 12051, 12052, 12052, 12052, 10825, 10826,
  10826, 10826, 8814, 8814, 8814, 8814, 6499, 6500, 6500, 6500, 4427, 4427,
  4427, 4427, 2732, 2733, 2733, 2733, 1557, 1557, 1557, 1557, 823, 823, 823,
  823, 414, 415, 415, 415, 213, 213, 213, 213, 116, 116, 116, 116, 79, 79, 79,
  79, 69, 69, 70, 70, 65, 66, 66, 66, 66, 66, 66, 66, 60, 61, 61, 61, 60, 61,
  61, 61, 64, 64, 64, 64, 58, 58, 58, 58, 57, 57, 57, 57, 65, 65, 65, 65, 63,
  63, 63, 63, 63, 63, 64, 64, 67, 67, 68, 68, 62, 62, 62, 62, 51, 51, 51, 51,
  57, 57, 58, 58, 50, 50, 50, 50, 52, 53, 53, 53, 47, 47, 48, 48, 57, 57, 57,
  57, 51, 52, 52, 52, 53, 53, 53, 53, 53, 53, 53, 53, 48, 49, 49, 49, 46, 46,
  47, 47, 47, 48, 48, 48, 56, 56, 56, 56, 57, 57, 58, 58, 50, 50, 50, 50, 59,
  59, 60, 60, 52, 52, 52, 52, 59, 59, 60, 60, 49, 49, 50, 50, 52, 53, 53, 53,
  56, 56, 56, 56, 54, 54, 55, 55, 48, 48, 49, 49, 58, 58, 58, 58, 53, 53, 53,
  53, 60, 60, 61, 61, 64, 64, 64, 64, 69, 69, 69, 69, 73, 73, 74, 74, 73, 73,
  73, 73, 95, 95, 95, 95, 91, 92, 92, 92, 87, 87, 88, 88, 92, 92, 92, 92, 90,
  90, 90, 90, 78, 78, 78, 78, 65, 65, 66, 66, 66, 66, 66, 66, 58, 58, 58, 58,
  50, 50, 50, 50, 50, 50, 50, 50, 59, 59, 59, 59, 50, 50, 50, 50, 48, 48, 48,
  48, 49, 49, 49, 49, 49, 50, 50, 50, 53, 53, 53, 53, 48, 48, 48, 48, 52, 52,
  52, 52, 53, 53, 53, 53, 52, 53, 53, 53, 60, 60, 60, 60, 54, 54, 54, 54, 58,
  58, 59, 59, 46, 46, 47, 47, 57, 57, 57, 57, 60, 60, 61, 61, 55, 55, 56, 56,
  57, 57, 57, 57, 58, 58, 58, 58, 63, 63, 64, 64, 70, 70, 70, 70, 67, 67, 67,
  67, 66, 67, 67, 67, 74, 74, 74, 74, 77, 77, 77, 77, 83, 84, 84, 84, 84, 85,
  85, 85, 86, 86, 86, 86, 86, 86, 86, 86, 81, 81, 81, 81, 81, 81, 81, 81, 85,
  85, 85, 85, 83, 83, 83, 83, 73, 74, 74, 74, 76, 76, 77, 77, 72, 73, 73, 73,
  72, 73, 73, 73, 72, 72, 72, 72, 70, 70, 70, 70, 72, 72, 72, 72, 73, 73, 73,
  73, 75, 76, 76, 76, 79, 79, 80, 80, 88, 88, 89, 89, 89, 89, 89, 89, 86, 87,
  87, 87, 94, 94, 95, 95, 109, 109, 110, 110, 108, 108, 108, 108, 117, 117,
  117, 117, 136, 136, 136, 136, 140, 140, 141, 141, 143, 143, 143, 143, 147,
  147, 147, 147, 142, 143, 143, 143, 144, 144, 144, 144, 142, 142, 142, 142,
  148, 149, 149, 149, 148, 148, 148, 148, 144, 144, 144, 144, 151, 151, 152,
  152, 168, 169, 169, 169, 158, 158, 158, 158, 177, 177, 177, 177, 205, 205,
  206, 206, 227, 227, 228, 228, 303, 304, 304, 304, 417, 417, 417, 417, 646,
  646, 646, 646, 1008, 1008, 1009, 1009, 1612, 1612, 1612, 1612, 2540, 2541,
  2541, 2541, 3889, 3889, 3889, 3889, 5630, 5630, 5631, 5631, 7687, 7687,
  7687, 7687, 9731, 9731, 9731, 9731, 11641, 11641, 11641, 11641, 12759,
  12759, 12759, 12759, 13176, 13177, 13177, 13177, 12544, 12544, 12544, 12544,
  11201, 11202, 11202, 11202, 9218, 9218, 9219, 9219, 7234, 7234, 7234, 7234,
  5113, 5113, 5113, 5113, 3470, 3470, 3470, 3470, 2219, 2219, 2219, 2219,
  1372, 1373, 1373, 1373, 809, 810, 810, 810, 512, 513, 513, 513, 345, 345,
  345, 345, 280, 280, 280, 280, 240, 240, 241, 241, 239, 239, 239, 239, 237,
  238, 238, 238, 277, 278, 278, 278, 277, 277, 277, 277, 294, 294, 294, 294,
  315, 316, 316, 316, 332, 333, 333, 333, 372, 373, 373, 373, 393, 394, 394,
  394, 433, 433, 433, 433, 472, 472, 473, 473, 524, 525, 525, 525, 577, 578,
  578, 578, 697, 697, 697, 697, 842, 842, 843, 843, 1050, 1050, 1050, 1050,
  1274, 1274, 1275, 1275, 1560, 1560, 1561, 1561, 1882, 1882, 1883, 1883,
  2173, 2173, 2174, 2174, 2432, 2432, 2432, 2432, 2567, 2567, 2567, 2567,
  2556, 2556, 2556, 2556, 2409, 2409, 2409, 2409, 2155, 2155, 2155, 2155,
  1813, 1814, 1814, 1814, 1392, 1393, 1393, 1393, 1098, 1098, 1098, 1098, 806,
  807, 807, 807, 601, 601, 602, 602, 435, 436, 436, 436, 399, 399, 400, 400,
  389, 389, 389, 389, 444, 444, 444, 444, 508, 509, 509, 509, 653, 653, 653,
  653, 810, 810, 810, 810, 1001, 1002, 1002, 1002, 1276, 1276, 1276, 1276,
  1585, 1585, 1586, 1586, 1970, 1970, 1970, 1970, 2415, 2415, 2416, 2416,
  3031, 3031, 3032, 3032, 3750, 3751, 3751, 3751, 4620, 4620, 4620, 4620,
  5435, 5436, 5436, 5436, 6371, 6371, 6371, 6371, 7091, 7091, 7091, 7091,
  7777, 7777, 7778, 7778, 8011, 8011, 8011, 8011, 7893, 7893, 7893, 7893,
  7455, 7455, 7455, 7455, 6712, 6712, 6713, 6713, 5690, 5690, 5690, 5690,
  4601, 4601, 4601, 4601, 3557, 3557, 3558, 3558, 2642, 2643, 2643, 2643,
  1821, 1821, 1821, 1821, 1231, 1231, 1232, 1232, 792, 792, 792, 792, 474,
  474, 475, 475, 277, 278, 278, 278, 163, 163, 163, 163, 86, 87, 87, 87, 50,
  50, 51, 51, 33, 33, 33, 33, 17, 17, 17, 17, 6, 7, 7, 7, 11, 12, 12, 12, 8,
  8, 8, 8, 5, 5, 5, 5, 6, 6, 6, 6, 5, 6, 6, 6, 7, 7, 7, 7, 2, 3, 3, 3, 4, 4,
  4, 4, 2, 2, 3, 3, 1, 1, 2, 2, 1, 2, 2, 2, 3, 3, 4, 4, 3, 3, 3, 3, 3, 4, 4,
  4, 3, 3, 3, 3, 4, 4, 4, 4, 2, 3, 3, 3, 5, 5, 5, 5, 2, 2, 2, 2, 4, 4, 4, 4,
  2, 2, 2, 2, 3, 4, 4, 4, 2, 3, 3, 3, 2, 2, 2, 2, 3, 4, 4, 4, 2, 2, 2, 2, 2,
  3, 3, 3, 2, 2, 3, 3, 2, 2, 3, 3, 1, 1, 1, 1, 3, 3, 3, 3, 3, 3, 3, 3, 2, 3,
  3, 3, 2, 3, 3, 3, 2, 2, 2, 2, 1, 1, 1, 1, 2, 3, 3, 3, 2, 2, 2, 2, 1, 2, 2,
  2, 2, 2, 3, 3, 1, 1, 2, 2, 3, 3, 3, 3, 2, 2, 2, 2, 1, 2, 2, 2, 4, 4, 4, 4,
  2, 2, 2, 2, 2, 2, 2, 2, 2, 3, 3, 3, 1, 1, 2, 2, 2, 3, 3, 3, 1, 2, 2, 2, 1,
  1, 1, 1, 1, 1, 1, 1, 1, 2, 2, 2, 1, 1, 1, 1, 0, 1, 1, 1, 1, 1, 1, 1, 1, 1,
  1, 1, 2, 2, 2, 2, 0, 1, 1, 1, 1, 1, 1, 1, 1, 1, 1, 1, 2, 2, 3, 3, 1, 1, 1,
  1, 2, 2, 2, 2, 2, 3, 3, 3, 1, 2, 2, 2, 2, 2, 2, 2, 2, 2, 2, 2, 2, 2, 2, 2,
  0, 1, 1, 1, 3, 3, 4, 4, 1, 1, 1, 1, 2, 2, 2, 2, 2, 3, 3, 3, 2, 2, 2, 2, 2,
  2, 2, 2, 1, 2, 2, 2, 2, 2, 2, 2, 1, 2, 2, 2, 2, 2, 2, 2, 1, 2, 2, 2, 1, 1,
  1, 1, 3, 3, 3, 3, 2, 2, 3, 3, 2, 3, 3, 3, 4, 4, 4, 4, 2, 2, 2, 2, 2, 2, 2,
  2, 2, 2, 3, 3, 2, 2, 2, 2, 4, 4, 4, 4, 3, 3, 4, 4, 6, 6, 7, 7, 6, 6, 7, 7,
  9, 9, 9, 9, 11, 11, 12, 12, 10, 10, 10, 10, 16, 16, 17, 17, 13, 13, 14, 14,
  15, 15, 15, 15, 18, 18, 18, 18, 16, 16, 17, 17, 17, 17, 17, 17, 13, 13, 13,
  13, 13, 13, 13, 13, 11, 11, 11, 11, 9, 9, 9, 9, 7, 7, 7, 7, 9, 9, 9, 9, 4,
  5, 5, 5, 5, 6, 6, 6, 8, 8, 9, 9, 8, 8, 8, 8, 9, 9, 10, 10, 8, 9, 9, 9, 14,
  14, 15, 15, 14, 15, 15, 15, 11, 11, 12, 12, 9, 9, 10, 10, 9, 9, 9, 9, 7, 7,
  8, 8, 7, 7, 7, 7, 6, 6, 6, 6, 4, 4, 4, 4, 5, 6, 6, 6, 3, 3, 4, 4, 2, 3, 3,
  3, 2, 3, 3, 3, 3, 4, 4, 4, 4, 4, 4, 4, 4, 4, 4, 4, 4, 4, 4, 4, 4, 5, 5, 5,
  7, 7, 7, 7, 6, 6, 6, 6, 5, 6, 6, 6, 5, 5, 6, 6, 7, 8, 8, 8, 4, 5, 5, 5, 3,
  4, 4, 4, 5, 5, 5, 5, 3, 3, 3, 3, 3, 3, 3, 3, 3, 3, 3, 3, 2, 3, 3, 3, 4, 5,
  5, 5, 5, 5, 5, 5, 4, 4, 4, 4, 4, 4, 4, 4, 5, 5, 6, 6, 4, 5, 5, 5, 5, 6, 6,
  6, 4, 4, 4, 4, 7, 7, 8, 8, 7, 7, 7, 7, 7, 7, 8, 8, 9, 10, 10, 10, 7, 8, 8,
  8, 12, 13, 13, 13, 8, 8, 9, 9, 8, 8, 9, 9, 11, 12, 12, 12, 12, 13, 13, 13,
  13, 13, 14, 14, 13, 13, 14, 14, 13, 14, 14, 14, 11, 12, 12, 12, 8, 8, 9, 9,
  12, 12, 13, 13, 8, 8, 8, 8, 10, 10, 10, 10, 9, 10, 10, 10, 8, 8, 8, 8, 5, 5,
  5, 5, 8, 8, 8, 8, 9, 9, 9, 9, 8, 9, 9, 9, 10, 10, 10, 10, 9, 9, 10, 10, 11,
  11, 11, 11, 12, 12, 12, 12, 8, 9, 9, 9, 7, 7, 8, 8, 8, 9, 9, 9, 7, 7, 7, 7,
  7, 7, 8, 8, 7, 7, 8, 8, 6, 6, 6, 6, 6, 6, 6, 6, 7, 7, 7, 7, 7, 7, 7, 7, 6,
  6, 6, 6, 5, 5, 6, 6, 5, 6, 6, 6, 6, 6, 6, 6, 3, 3, 3, 3, 1, 2, 2, 2, 3, 4,
  4, 4, 3, 3, 4, 4, 4, 4, 4, 4, 2, 2, 3, 3, 3, 3, 3, 3, 2, 2, 3, 3, 3, 3, 3,
  3, 5, 5, 6, 6, 4, 4, 4, 4, 5, 5, 5, 5, 3, 3, 3, 3, 2, 3, 3, 3, 4, 4, 5, 5,
  2, 2, 3, 3, 4, 4, 5, 5, 4, 4, 5, 5, 3, 3, 4, 4, 5, 5, 5, 5, 5, 5, 5, 5, 3,
  4, 4, 4, 4, 4, 5, 5, 6, 6, 7, 7, 5, 5, 5, 5, 5, 6, 6, 6, 6, 6, 6, 6, 7, 7,
  7, 7, 7, 7, 7, 7, 6, 6, 6, 6, 6, 7, 7, 7, 8, 9, 9, 9, 6, 7, 7, 7, 7, 7, 7,
  7, 6, 6, 6, 6, 6, 6, 7, 7, 8, 8, 9, 9, 11, 11, 11, 11, 13, 14, 14, 14, 11,
  12, 12, 12, 8, 8, 9, 9, 15, 15, 15, 15, 11, 12, 12, 12, 15, 15, 15, 15, 12,
  13, 13, 13, 15, 16, 16, 16, 12, 12, 13, 13, 11, 12, 12, 12, 13, 13, 13, 13,
  16, 17, 17, 17, 18, 18, 19, 19, 17, 17, 18, 18, 17, 17, 17, 17, 13, 13, 14,
  14, 21, 21, 21, 21, 16, 16, 17, 17, 15, 16, 16, 16, 19, 19, 19, 19, 21, 22,
  22, 22, 20, 20, 20, 20, 19, 19, 19, 19, 23, 23, 24, 24, 26, 26, 26, 26, 26,
  26, 26, 26, 33, 33, 33, 33, 45, 45, 46, 46, 58, 58, 58, 58, 83, 83, 84, 84,
  119, 119, 120, 120, 174, 175, 175, 175, 253, 254, 254, 254, 344, 344, 344,
  344, 472, 472, 473, 473, 645, 645, 645, 645, 842, 842, 842, 842, 1107, 1107,
  1108, 1108, 1280, 1280, 1280, 1280, 1504, 1505, 1505, 1505, 1642, 1643,
  1643, 1643, 1802, 1802, 1802, 1802, 1870, 1870, 1871, 1871, 1813, 1814,
  1814, 1814, 1708, 1708, 1709, 1709, 1519, 1519, 1520, 1520, 1332, 1332,
  1332, 1332, 1110, 1111, 1111, 1111, 873, 873, 874, 874, 708, 709, 709, 709,
  512, 512, 513, 513, 364, 364, 364, 364, 243, 243, 243, 243, 172, 172, 172,
  172, 117, 117, 118, 118, 75, 75, 75, 75, 56, 57, 57, 57, 40, 40, 40, 40, 32,
  33, 33, 33, 34, 34, 34, 34, 33, 33, 33, 33, 29, 29, 30, 30, 32, 32, 32, 32,
  35, 35, 36, 36, 30, 31, 31, 31, 35, 35, 36, 36, 33, 33, 34, 34, 41, 42, 42,
  42, 37, 37, 37, 37, 42, 42, 42, 42, 42, 42, 43, 43, 49, 49, 49, 49, 48, 48,
  48, 48, 51, 51, 51, 51, 47, 47, 47, 47, 49, 49, 49, 49, 57, 57, 57, 57, 57,
  57, 58, 58, 62, 62, 63, 63, 58, 58, 58, 58, 57, 57, 58, 58, 67, 67, 68, 68,
  69, 69, 69, 69, 68, 69, 69, 69, 82, 82, 82, 82, 79, 80, 80, 80, 84, 85, 85,
  85, 90, 90, 90, 90, 101, 101, 101, 101, 100, 101, 101, 101, 112, 113, 113,
  113, 119, 119, 120, 120, 121, 121, 122, 122, 125, 125, 125, 125, 130, 130,
  130, 130, 144, 145, 145, 145, 161, 162, 162, 162, 164, 165, 165, 165, 181,
  182, 182, 182, 197, 197, 197, 197, 204, 204, 205, 205, 214, 214, 215, 215,
  241, 241, 242, 242, 266, 266, 266, 266, 298, 298, 298, 298, 304, 304, 305,
  305, 341, 342, 342, 342, 352, 352, 353, 353, 397, 398, 398, 398, 424, 425,
  425, 425, 455, 455, 455, 455, 504, 505, 505, 505, 520, 521, 521, 521, 522,
  523, 523, 523, 557, 557, 558, 558, 598, 598, 598, 598, 639, 639, 639, 639,
  650, 650, 651, 651, 656, 656, 656, 656, 687, 688, 688, 688, 749, 750, 750,
  750, 720, 720, 720, 720, 750, 750, 750, 750, 745, 746, 746, 746, 760, 760,
  760, 760, 751, 751, 751, 751, 717, 717, 717, 717, 696, 696, 696, 696, 652,
  652, 652, 652, 593, 593, 593, 593, 537, 537, 538, 538, 493, 493, 493, 493,
  476, 477, 477, 477, 433, 434, 434, 434, 427, 428, 428, 428, 426, 426, 427,
  427, 463, 463, 463, 463, 512, 512, 513, 513, 567, 567, 568, 568, 662, 662,
  662, 662, 705, 705, 705, 705, 810, 810, 811, 811, 845, 845, 846, 846, 906,
  906, 906, 906, 942, 942, 942, 942, 923, 923, 923, 923, 857, 857, 857, 857,
  786, 787, 787, 787, 703, 703, 703, 703, 622, 622, 622, 622, 498, 499, 499,
  499, 402, 402, 402, 402, 314, 315, 315, 315, 236, 236, 236, 236, 159, 159,
  160, 160, 118, 118, 118, 118, 76, 76, 76, 76, 54, 54, 54, 54, 32, 32, 32,
  32, 19, 20, 20, 20, 12, 12, 13, 13, 7, 8, 8, 8, 4, 5, 5, 5, 4, 5, 5, 5, 2,
  2, 2, 2, 1, 1, 2, 2, 1, 1, 1, 1, 1, 1, 1, 1, 0, 0, 1, 1, 0, 0, 0, 0, 0, 1,
  1, 1, 0, 0, 0, 0, 0, 0, 1, 1, 0, 0, 1, 1, 0, 0, 0, 0, 1, 1, 1, 1, 0, 1, 1,
  1, 0, 0, 1, 1, 0, 0, 1, 1, 0, 1, 1, 1, 0, 0, 1, 1, 0, 0, 0, 0, 0, 0, 1, 1,
  0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 1, 1, 1, 0, 0, 0, 0, 0, 0, 1, 1, 0,
  0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0,
  0, 0, 0, 0, 1, 1, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0,
  0, 0, 0, 0, 0, 0, 0, 1, 1, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 1, 1, 0, 0, 0, 0,
  0, 0, 0, 0, 0, 0, 0, 0, 0, 1, 1, 1, 0, 0, 1, 1, 0, 0, 1, 1, 0, 0, 0, 0, 0,
  0, 1, 1, 0, 0, 1, 1, 0, 0, 0, 0, 1, 2, 2, 2, 1, 1, 2, 2, 0, 0, 1, 1, 1, 1,
  1, 1, 1, 2, 2, 2, 3, 3, 4, 4, 1, 2, 2, 2, 3, 3, 3, 3, 2, 3, 3, 3, 3, 3, 3,
  3, 1, 1, 2, 2, 2, 2, 2, 2, 1, 1, 2, 2, 1, 1, 2, 2, 1, 2, 2, 2, 1, 1, 1, 1,
  1, 1, 2, 2, 0, 1, 1, 1, 0, 0, 1, 1, 1, 1, 1, 1, 1, 2, 2, 2, 1, 1, 1, 1, 1,
  1, 2, 2, 0, 0, 1, 1, 1, 1, 1, 1, 1, 1, 1, 1, 1, 2, 2, 2, 1, 1, 2, 2, 0, 1,
  1, 1, 0, 0, 1, 1, 1, 1, 1, 1, 1, 1, 1, 1, 1, 1, 1, 1, 0, 0, 0, 0, 0, 1, 1,
  1, 0, 0, 1, 1, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 1, 1,
  0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0,
  0, 0, 0, 0, 0, 0, 0, 0, 0, 1, 1, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0,
  0, 0, 0, 0, 0, 0, 0, 1, 1, 1, 0, 0, 1, 1, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 1,
  1, 0, 0, 1, 1, 0, 0, 1, 1, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 1, 1, 1,
  0, 0, 0, 0, 0, 0, 1, 1, 0, 0, 0, 0, 0, 0, 1, 1, 1, 1, 1, 1, 0, 0, 1, 1, 1,
  1, 1, 1, 1, 1, 1, 1, 0, 0, 1, 1, 0, 1, 1, 1, 0, 1, 1, 1, 0, 1, 1, 1, 1, 1,
  1, 1, 1, 1, 2, 2, 0, 0, 0, 0, 1, 1, 1, 1, 1, 1, 1, 1, 1, 1, 1, 1, 0, 0, 1,
  1, 1, 1, 1, 1, 0, 1, 1, 1, 2, 2, 2, 2, 1, 1, 1, 1, 1, 1, 1, 1, 0, 1, 1, 1,
  0, 0, 1, 1, 1, 1, 1, 1, 0, 0, 0, 0, 1, 2, 2, 2, 1, 1, 1, 1, 1, 1, 2, 2, 0,
  1, 1, 1, 1, 1, 1, 1, 0, 1, 1, 1, 1, 1, 1, 1, 1, 2, 2, 2, 2, 2, 2, 2, 0, 1,
  1, 1, 1, 1, 1, 1, 2, 3, 3, 3, 2, 2, 3, 3, 2, 2, 3, 3, 1, 1, 1, 1, 2, 2, 2,
  2, 2, 2, 3, 3, 1, 2, 2, 2, 1, 1, 2, 2, 1, 1, 1, 1, 0, 0, 1, 1, 1, 1, 1, 1,
  1, 1, 1, 1, 0, 1, 1, 1, 1, 1, 1, 1, 1, 1, 2, 2, 1, 1, 2, 2, 0, 0, 0, 0, 0,
  1, 1, 1
]

if __name__ == '__main__':
  try:
    initializeSignalHandlers()
    main(sys.argv)
  except Exception, e:
    if isinstance(e, SystemExit):
      raise e
    else:
      print >> sys.stderr
      print >> sys.stderr, "Error: %s" % e
      sys.exit(1)
